# -*- coding: utf-8 -*-
"""Helper methods for CSDM class."""
from __future__ import print_function

from copy import deepcopy

import numpy as np
from astropy.units.quantity import Quantity

from .units import ScalarQuantity

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

__all__ = ["LITERALS_ENCODING", "QuantityType", "NumericType"]

literals_quantity_type = {
    "scalar": lambda n: 1,
    "vector": lambda n: n,
    "matrix": lambda n: n,
    "audio": lambda n: n,
    "pixel": lambda n: n,
    "symmetric_matrix": lambda n: int(n * (n + 1) / 2),
}

LITERALS_ENCODING = ("base64", "none", "raw")
literals_quantity_type_ = [
    "scalar",
    "vector_n",
    "matrix_n_m",
    # "audio_n",
    "pixel_n",
    "symmetric_matrix_n",
]

numpy_scalars = (
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.float32,
    np.float64,
    np.complex64,
    np.complex128,
)


class QuantityType:
    """Validate the quantity_type string value.

    The valid options are `scalar`, `vector_n`, `pixel_n`, `matrix_n_m`, and
    `symmetric_matrix_n`.

    :returns: The quantity_type key-value, if the value is valid.
    :raises KeyError: Otherwise.
    """

    __slots__ = ("value", "p")

    def __init__(self, element):
        r"""Instantiate a QuantityType class instance."""
        self.update(element)

    def __str__(self):
        r"""Return a string with the quantity type."""
        return self.value

    def __eq__(self, other):
        """Overrides the default implementation"""
        check = [self.value == other.value, self.p == other.p]
        return False if False in check else True

    def update(self, element):
        """Update the quantity type."""
        validate(element, "quantity_type", str, self._check_quantity_type)

    @classmethod
    def _get_number_of_components(cls, keyword, numbers):
        return int(literals_quantity_type[keyword](numbers.prod()))

    def _check_quantity_type(self, element):
        list_values = element.strip().split("_")
        numbers = np.asarray([int(item) for item in list_values if item.isnumeric()])
        keyword = "_".join([item for item in list_values if not item.isnumeric()])

        lst = literals_quantity_type
        if keyword not in lst:
            message = (
                f"The value, `{keyword}`, is an invalid `quantity_type` enumeration "
                f"literal. The allowed values are {literals_quantity_type_}."
            )
            raise ValueError(message)

        components = self._get_number_of_components(keyword, numbers)
        self.value = element
        self.p = components


class NumericType:
    """Validate the numeric_type string value.

    The valid options are `uint8`, `uint16`, `uint32`, `uint64`, `int8`, `int16`,
    `int32`, `int64`, `float32`, `float64`, `complex64`, and `complex128`. The byte
    order for multi-byte numeric_types are assumed to follow the little endianness
    format.

    :returns: The numeric_type value, if the value is valid.
    :raises KeyError: Otherwise.
    """

    __slots__ = ("value", "dtype")

    _lst = {
        "uint8": "<u1",
        "uint16": "<u2",
        "uint32": "<u4",
        "uint64": "<u8",
        "int8": "<i1",
        "int16": "<i2",
        "int32": "<i4",
        "int64": "<i8",
        # "float16": "<f2",
        "float32": "<f4",
        "float64": "<f8",
        "complex64": "<c8",
        "complex128": "<c16",
        ">u1": "<u1",
        ">u2": "<u2",
        ">u4": "<u4",
        ">u8": "<u8",
        ">i1": "<i1",
        ">i2": "<i2",
        ">i4": "<i4",
        ">i8": "<i8",
        # ">f2": "<f2",
        ">f4": "<f4",
        ">f8": "<f8",
        ">c8": "<c8",
        ">c16": "<c16",
    }

    literals = (
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "int8",
        "int16",
        "int32",
        "int64",
        # "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    )

    def __init__(self, element="float32"):
        """Instantiate a NumericType class instance."""
        self.update(element)

    def update(self, element):
        """Update the numeric type."""
        validate(
            element, "numeric_type", (str, type, np.dtype), self._check_numeric_type
        )

    def __str__(self):
        """Return a string with the numeric type."""
        return self.value

    def __eq__(self, other):
        """Overrides the default implementation"""
        check = [self.value == other.value, self.dtype == other.dtype]
        return False if False in check else True

    def _check_numeric_type(self, element):
        if isinstance(element, np.dtype):
            self.dtype = element
            self.value = str(element)
            return

        if isinstance(element, type):
            self.dtype = np.dtype(element)
            self.value = str(self.dtype)
            return

        lst = self.__class__._lst
        if element not in lst.keys():
            literals = self.__class__.literals
            message = (
                f"The value, `{element}`, is an invalid `numeric_type` enumeration "
                f"literal. The allowed values are {literals}."
            )
            raise ValueError(message)

        self.value = element
        self.dtype = np.dtype(lst[element])


def validate(value, attr, types, method=None):
    if isinstance(value, types):
        if method is None:
            return value
        return method(value)
    raise TypeError(type_error(types, attr, value))


def type_error(types, attr, value):
    types = (
        " or ".join([item.__name__ for item in types])
        if isinstance(types, tuple)
        else types.__name__
    )
    val_type = type(value).__name__
    return f"Expecting an instance of type `{types}` for {attr}, got `{val_type}`."


def _axis_label(label, unit=None):
    """Return a formatted label with units."""
    return f"{label} / ({unit})" if unit not in [None, ""] else label


def _get_dictionary(*arg, **kwargs):
    if arg != ():
        if isinstance(arg[0], dict):
            return arg[0]

        raise Exception(
            "The argument is either a dictionary with the allowed keywords or a "
            "collection of valid arguments."
        )
    return kwargs


def check_encoding(element):
    """Validate the encoding string value.

    The valid options are `base64`, `none`, and `raw`.

    :returns: The encoding key-value, if the value is valid.
    :raises KeyError: Otherwise.
    """
    if element in LITERALS_ENCODING:
        return element

    message = (
        f"The value, `{element}`, is an invalid `encoding` enumeration literal. "
        f"The allowed values are '{LITERALS_ENCODING}'."
    )
    raise ValueError(message)


def numpy_dtype_to_numeric_type(element):
    """Return a valid numeric_type value based on the dtype of numpy array."""
    lst = {
        "<u1": "uint8",
        "<u2": "uint16",
        "<u4": "uint32",
        "<u8": "uint64",
        "<i1": "int8",
        "<i2": "int16",
        "<i4": "int32",
        "<i8": "int64",
        # "<f2": "float16",
        "<f4": "float32",
        "<f8": "float64",
        "<c8": "complex64",
        "<c16": "complex128",
        ">u1": "uint8",
        ">u2": "uint16",
        ">u4": "uint32",
        ">u8": "uint64",
        ">i1": "int8",
        ">i2": "int16",
        ">i4": "int32",
        ">i8": "int64",
        # ">f2": "float16",
        ">f4": "float32",
        ">f8": "float64",
        ">c8": "complex64",
        ">c16": "complex128",
    }

    lst2 = (
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "int8",
        "int16",
        "int32",
        "int64",
        # "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    )

    if element in lst:
        return lst[element]
    if element in lst2:
        return element
    raise ValueError(f"The dtype, {element}, is not supported.")


def check_and_assign_bool(element):
    if element is None:
        return False
    return validate(element, "boolean", bool)


def check_scalar_object(other, operator=""):
    """Check if the object is scalar: int, float, complex, np.ndarray, Quantity, or
    ScalarQuantity.

    Returns: The other object.
    """
    if isinstance(other, numpy_scalars):
        return other

    if not isinstance(
        other, (int, float, complex, np.ndarray, Quantity, ScalarQuantity)
    ):
        raise TypeError(
            f"unsupported operand type(s) {operator}: 'CSDM' and "
            f"'{other.__class__.__name__}'."
        )
    if isinstance(other, np.ndarray):
        if other.ndim != 0:
            raise ValueError("Only scalar multiplication or division is allowed.")

    if isinstance(other, ScalarQuantity):
        return other.quantity

    return other


def _get_broadcast_shape(array, ndim, axis):
    """Return the broadcast array for numpy ndarray operations."""
    s = [None for i in range(ndim)]
    s[axis] = slice(None, None, None)
    return array[tuple(s)]


def _check_dimension_indices(d, index=-1):
    """Check the list of indexes to ensure that each index is an integer
    and within the counts of dimensions.
    """
    index = deepcopy(index)

    def _correct_index(i, d):
        if not isinstance(i, int):
            raise TypeError(f"{message}, found {type(i)}")
        if i < 0:
            i += d
        if i > d:
            raise IndexError(
                f"The `index` {index} cannot be greater than the total number of "
                f"dimensions - 1, {d}."
            )
        return -1 - i

    message = "Index/Indices are expected as integer(s)"
    if isinstance(index, tuple):
        index = list(index)
    if isinstance(index, (list, np.ndarray)):
        for i, item in enumerate(index):
            index[i] = _correct_index(item, d)
        return tuple(index)
    if isinstance(index, int):
        return tuple([_correct_index(index, d)])
    raise TypeError(f"{message}, found {type(index)}")



# -*- coding: utf-8 -*-
import numpy as np
import pytest

import csdmpy as cp


def test_csdm_new():
    a = cp.CSDM(dimensions=[cp.as_dimension(np.arange(10))])

    error = "Expecting a Dimension object"
    with pytest.raises(ValueError, match=f".*{error}.*"):
        a.dimensions.append(np.arange(5))

    with pytest.raises(ValueError, match=f".*{error}.*"):
        a.dimensions[0] = np.arange(5)

    error = "Deleting items is not allowed"
    with pytest.raises(LookupError, match=f".*{error}.*"):
        del a.dimensions[0]

    assert a.dimensions[0] != np.arange(5)
    assert a.ndim == 1
    assert a.to_dict() == {
        "csdm": {
            "dimensions": [{"count": 10, "increment": "1.0", "type": "linear"}],
            "version": "1.0",
        }
    }


def test_bad_csdm():
    error = "A list of valid Dimension objects or equivalent dictionary"
    with pytest.raises(ValueError, match=f".*{error}.*"):
        cp.CSDM(dimensions="blah")

    error = "A list of valid DependentVariable objects or equivalent dictionary"
    with pytest.raises(ValueError, match=f".*{error}.*"):
        cp.CSDM(dependent_variables="blah")



# -*- coding: utf-8 -*-
"""The LinearDimension sub type class."""
from __future__ import division
from __future__ import print_function

import numpy as np
from astropy.units import Quantity

from .base import _copy_core_metadata
from .quantitative import BaseQuantitativeDimension
from .quantitative import ReciprocalDimension
from csdmpy.units import frequency_ratio
from csdmpy.units import ScalarQuantity
from csdmpy.utils import check_and_assign_bool
from csdmpy.utils import check_scalar_object
from csdmpy.utils import validate

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["LinearDimension"]

# =========================================================================== #
#                          LinearDimension Class                              #
# =========================================================================== #


class LinearDimension(BaseQuantitativeDimension):
    """LinearDimension class.

    Generates an object representing a physical dimension whose coordinates are
    uniformly sampled along a grid dimension. See :ref:`linearDimension_uml` for
    details.
    """

    __slots__ = ("_count", "_increment", "_complex_fft", "reciprocal", "_coordinates")

    _type = "linear"

    def __init__(self, count, increment, complex_fft=False, **kwargs):
        """Instantiate a DimensionWithLinearSpacing class instance."""
        self._count = count
        self._increment = ScalarQuantity(increment).quantity
        self._complex_fft = check_and_assign_bool(complex_fft)
        _unit = self._increment.unit
        if "reciprocal" not in kwargs.keys():
            kwargs["reciprocal"] = {
                "increment": None,
                "coordinates_offset": None,
                "origin_offset": None,
                "period": None,
                "quantity_name": None,
                "label": "",
                "description": "",
                "application": {},
            }

        super().__init__(unit=_unit, **kwargs)

        # create a reciprocal dimension
        r_unit = self._unit ** -1
        self.reciprocal = ReciprocalDimension(unit=r_unit, **kwargs["reciprocal"])
        self._get_coordinates()

    def __repr__(self):
        meta = [f"{k}={v}" for k, v in self.dict().items() if k != "type"]
        properties = ", ".join(meta)
        return f"LinearDimension({properties})"

    def __str__(self):
        return f"LinearDimension({self.coordinates.__str__()})"

    def __eq__(self, other):
        """Overrides the default implementation"""
        other = other.subtype if hasattr(other, "subtype") else other
        if not isinstance(other, LinearDimension):
            return False
        check = [getattr(self, _) == getattr(other, _) for _ in __class__.__slots__[:4]]
        check += [super().__eq__(other)]
        return np.all(check)

    def __mul__(self, other):
        """Multiply the LinearDimension object by a scalar."""
        return _update_linear_dimension_object_by_scalar(self.copy(), other, "mul")

    def __imul__(self, other):
        """Multiply the LinearDimension object by a scalar, in-place."""
        return _update_linear_dimension_object_by_scalar(self, other, "mul")

    def __truediv__(self, other):
        """Divide the LinearDimension object by a scalar."""
        return _update_linear_dimension_object_by_scalar(self.copy(), other, "truediv")

    def __itruediv__(self, other):
        """Divide the LinearDimension object by a scalar, in-place."""
        return _update_linear_dimension_object_by_scalar(self, other, "truediv")

    def _swap(self):
        """Swap the value between the dimension and reciprocal dimension object."""
        attrs = [
            "_description",
            "_application",
            "_coordinates_offset",
            "_origin_offset",
            "_period",
            "_quantity_name",
            "_label",
            "_unit",
            "_equivalent_unit",
        ]
        for item in attrs:
            val = getattr(self, item)
            r_val = getattr(self.reciprocal, item)
            setattr(self, item, r_val)
            setattr(self.reciprocal, item, val)

    def _get_coordinates(self):
        index = np.arange(self._count, dtype=np.float64)
        index -= int(self._count / 2) if self._complex_fft else 0
        self._coordinates = index * self._increment.to(self._unit)

    # ----------------------------------------------------------------------- #
    #                                  Attributes                             #
    # ----------------------------------------------------------------------- #
    @property
    def type(self):
        """Return the type of the dimension."""
        return self.__class__._type

    @property
    def count(self):
        """Total number of points along the linear dimension."""
        return self._count

    @count.setter
    def count(self, value):
        value = validate(value, "count", int)
        self._count = value
        self._get_coordinates()

    @property
    def increment(self):
        """Increment along the linear dimension."""
        return self._increment

    @increment.setter
    def increment(self, value):
        allowed_types = (Quantity, str, ScalarQuantity)
        value = validate(value, "increment", allowed_types)
        self._increment = ScalarQuantity(value, self._unit).quantity
        self._get_coordinates()

    @property
    def complex_fft(self):
        """If True, orders the coordinates according to FFT output order."""
        return self._complex_fft

    @complex_fft.setter
    def complex_fft(self, value):
        self._complex_fft = validate(value, "complex_fft", bool)
        self._get_coordinates()

    @property
    def coordinates(self):
        """Return the coordinates along the dimensions."""
        n = self._count
        coordinates = self._coordinates[:n] + self.coordinates_offset

        equivalent_fn = self._equivalencies
        equivalent_unit = self._equivalent_unit

        if equivalent_fn is None or equivalent_unit is None:
            return coordinates.to(self._unit)

        if equivalent_fn == "nmr_frequency_ratio":
            denominator = self.origin_offset - self.get_nmr_reference_offset()
            if denominator.value == 0:
                raise ZeroDivisionError("Cannot convert the coordinates to ppm.")
            return coordinates.to(equivalent_unit, frequency_ratio(denominator))

        return coordinates.to(equivalent_unit, equivalent_fn)

    @coordinates.setter
    def coordinates(self, value):
        raise AttributeError(
            "The attribute cannot be modifed for Dimension objects with `linear` "
            "type. Use the `count`, `increment` or `coordinates_offset` attributes"
            " to update the coordinate along the linear dimension."
        )

    def get_nmr_reference_offset(self):
        """Calculate reference offset for NMR datsets."""
        if self.complex_fft:
            return self.coordinates_offset

        if self.count % 2 != 0:  # odd count
            return self.coordinates_offset + (self.count - 1) * self.increment / 2.0

        # even count
        n = self.count / 2
        if self.increment > 0:  # positive increment
            return self.coordinates_offset + n * self.increment

        # negative increment
        return self.coordinates_offset + (n - 1) * self.increment

    # ----------------------------------------------------------------------- #
    #                                 Methods                                 #
    # ----------------------------------------------------------------------- #

    def _copy_metadata(self, obj, copy=False):
        """Copy LinearDimension metadata."""
        obj = obj.subtype if hasattr(obj, "subtype") else obj
        if isinstance(obj, LinearDimension):
            _copy_core_metadata(self, obj, "linear")

    def dict(self):
        """Return the LinearDimension as a python dictionary."""
        obj = {}
        obj["type"] = self.__class__._type
        obj["count"] = self._count
        obj["increment"] = str(ScalarQuantity(self.increment))
        obj.update(super().dict())

        if self.complex_fft:
            obj["complex_fft"] = True

        reciprocal_obj = self.reciprocal.dict()
        if reciprocal_obj != {}:
            obj["reciprocal"] = reciprocal_obj

        return obj

    def reciprocal_coordinates(self):
        """Return reciprocal coordinates assuming Nyquist-shannan theorem."""
        coordinates_offset = self.reciprocal._coordinates_offset
        return self._reciprocal_coordinates() + coordinates_offset

    def _reciprocal_coordinates(self):
        """Return reciprocal coordinates assuming Nyquist-shannan theorem
        without the coordinates offset."""
        count = self._count
        increment = 1.0 / (count * self._increment)
        coordinates = np.arange(count) * increment
        return (
            coordinates
            if self.complex_fft
            else (coordinates - int(count / 2) * increment)
        )

    def reciprocal_increment(self):
        """Return reciprocal increment assuming Nyquist-shannan theorem."""
        return 1.0 / (self._count * self._increment)


def _update_linear_dimension_object_by_scalar(object_, other, type_="mul"):
    """Update object by multiplying by a scalar."""
    other = check_scalar_object(other)

    if type_ == "mul":
        object_._increment *= other
        object_._coordinates *= other
        object_._coordinates_offset *= other
        object_._origin_offset *= other
        object_._period *= other

    if type_ == "truediv":
        object_._increment /= other
        object_._coordinates /= other
        object_._coordinates_offset /= other
        object_._origin_offset /= other
        object_._period /= other

    object_._unit = object_._increment._unit
    object_._quantity_name = object_._unit.physical_type
    object_._equivalencies = None
    _reciprocal_unit = object_._unit ** -1
    object_.reciprocal = ReciprocalDimension(unit=_reciprocal_unit)
    return object_



# -*- coding: utf-8 -*-
"""The Internal DependentVariable SubType class."""
from __future__ import division
from __future__ import print_function

import numpy as np

from .base_class import BaseDependentVariable
from .decoder import Decoder
from csdmpy.utils import numpy_dtype_to_numeric_type


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["InternalDataset"]


class InternalDataset(BaseDependentVariable):
    """InternalDataset class."""

    def __init__(self, **kwargs):
        """Initialize."""
        components = kwargs["components"]
        if isinstance(components, list) and components != []:
            if isinstance(components[0], np.ndarray):
                components = np.asarray(components)

        if isinstance(components, np.ndarray):
            if kwargs["numeric_type"] is None:
                kwargs["numeric_type"] = numpy_dtype_to_numeric_type(
                    str(components.dtype)
                )
                kwargs["components"] = components
            else:
                kwargs["components"] = components.astype(kwargs["numeric_type"])

        if kwargs["numeric_type"] is None:
            raise KeyError(
                "Missing a required `numeric_type` key from the DependentVariable "
                "object."
            )

        # super base class must be initialized before retrieving
        # the components array.
        super().__init__(**kwargs)

        if not isinstance(components, np.ndarray):
            self._components = Decoder(
                self._encoding,
                self._quantity_type,
                components,
                self._numeric_type.dtype,
            )

        size = self._components.size
        p = self.quantity_type.p
        self._components.shape = (p, int(size / p))

    def dict(self, filename=None, dataset_index=None, for_display=False):
        """Return InternalDataset object as a python dictionary."""
        dictionary = {}
        dictionary["type"] = "internal"
        dictionary.update(super().dict(filename, dataset_index, for_display))
        return dictionary



# -*- coding: utf-8 -*-
import json
import warnings
from copy import deepcopy

from csdmpy.utils import validate


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"


class BaseDimension:
    """Base dimension class."""

    __slots__ = ("_label", "_description", "_application")

    def __init__(self, label, application, description):
        self._description = description
        self._application = application
        self._label = label

    def __eq__(self, other):
        """Check if two objects are equal"""
        check = [getattr(self, _) == getattr(other, _) for _ in __class__.__slots__]
        return False if False in check else True

    @property
    def label(self):
        """Label associated with the dimension."""
        return deepcopy(self._label)

    @label.setter
    def label(self, label=""):
        self._label = validate(label, "label", str)

    @property
    def application(self):
        """Return an application dimension associated with the dimensions."""
        return self._application

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def description(self):
        """Return a description of the dimension."""
        return deepcopy(self._description)

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)

    @property
    def coordinates(self):
        pass

    @coordinates.setter
    def coordinates(self, value):
        pass

    @property
    def coords(self):
        """Alias for the `coordinates` attribute."""
        return self.coordinates

    @coords.setter
    def coords(self, value):
        self.coordinates = value

    @property
    def axis_label(self):
        """Return a formatted string for displaying label along the dimension axis."""
        return self.label

    def to_dict(self):
        """Alias to the `dict()` method of the class."""
        return self.dict()

    def dict(self):
        """Return BaseDimension as a python dictionary"""
        obj = {}
        obj["label"] = self._label.strip()
        obj["description"] = self._description.strip()
        obj["application"] = self._application
        _ = [obj.pop(item) for item in [k for k, v in obj.items() if v in ["", {}]]]
        return obj

    def copy(self):
        """Return a copy of the object."""
        return deepcopy(self)

    @property
    def data_structure(self):
        """Json serialized string describing the Dimension object class instance."""
        return json.dumps(self.dict(), ensure_ascii=False, sort_keys=False, indent=2)


def check_count(value, total_count, dimension_type):
    """check the value for count."""
    value = validate(value, "count", int)
    if value > total_count:
        raise ValueError(
            f"Cannot set the count, {value}, more than the number of coordinates, "
            f"{total_count}, for the {dimension_type} dimensions."
        )

    if value < total_count:
        warnings.warn(f"The number of labels, {total_count}, are truncated to {value}.")

    return value


def _copy_core_metadata(obj1, obj2, dimension_type):
    """copy metadata from obj2 to obj1."""
    obj1._description = obj2._description
    obj1._application = obj2._application
    obj1._label = obj2._label

    if dimension_type == "labeled":
        return

    obj1._origin_offset = obj2._origin_offset
    obj1._period = obj2._period
    obj1.reciprocal = obj2.reciprocal
    obj1._equivalent_unit = obj2._equivalent_unit
    obj1._equivalencies = obj2._equivalencies

    if dimension_type == "monotonic":
        return

    obj1._complex_fft = obj2._complex_fft
    return



# -*- coding: utf-8 -*-
from collections.abc import MutableSequence

import numpy as np

from .dependent_variable import DependentVariable
from .dimension import Dimension
from .dimension import LabeledDimension
from .dimension import LinearDimension
from .dimension import MonotonicDimension

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"

__all__ = ["DimensionList", "DependentVariableList"]

__dimensions_list__ = (Dimension, LinearDimension, MonotonicDimension, LabeledDimension)


class AbstractList(MutableSequence):
    def __init__(self, data=[]):
        super().__init__()
        self._list = list(data)

    def __repr__(self):
        """String representation"""
        return self._list.__repr__()

    def __str__(self):
        """String representation"""
        string = ",\n".join([item.__repr__() for item in self._list])
        return f"[{string}]"

    def __len__(self):
        """List length"""
        return len(self._list)

    def __getitem__(self, index):
        """Get a list item"""
        return self._list[index]

    def __delitem__(self, index):
        raise LookupError("Deleting items is not allowed.")

    def check_object(self, *args):
        pass

    def insert(self, index: int, item: object):
        """Insert a list item"""
        item = self.check_object(item)
        self._list.insert(index, item)

    def append(self, item):
        """Append a list item"""
        item = self.check_object(item)
        self._list.append(item)

    def __setitem__(self, index, item):
        """Set item at index"""
        item = self.check_object(item)
        # if self._list[index].count != item.count:
        #     raise IndexError("Index out of range")
        self._list[index] = item

    def __eq__(self, other):
        """Check equality of DependentVariableList."""
        if not isinstance(other, self.__class__):
            return False
        if len(self._list) != len(other._list):
            return False

        check = [self_i == other_i for self_i, other_i in zip(self._list, other._list)]
        return np.all(check)


class DimensionList(AbstractList):
    def check_object(self, obj):
        if isinstance(obj, dict):
            obj = Dimension(**obj)
        if not isinstance(obj, __dimensions_list__):
            name = obj.__class__.__name__
            raise ValueError(f"Expecting a Dimension object, found {name}")
        return obj


class DependentVariableList(AbstractList):
    def check_object(self, obj):
        if isinstance(obj, dict):
            obj = DependentVariable(**obj)
        if not isinstance(obj, DependentVariable):
            name = obj.__class__.name__
            raise ValueError(f"Expecting a DependentVariable object, found {name}")
        return obj



# -*- coding: utf-8 -*-
"""Base Quantitative class."""
from copy import deepcopy

import numpy as np
from astropy.units import Quantity

from .base import BaseDimension
from csdmpy.units import check_quantity_name
from csdmpy.units import ScalarQuantity
from csdmpy.utils import _axis_label
from csdmpy.utils import type_error
from csdmpy.utils import validate

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["BaseQuantitativeDimension", "ReciprocalDimension"]

ALLOWED_TYPES = (Quantity, str, ScalarQuantity)
# =========================================================================== #
#                          Base Quantitative Class                            #
# =========================================================================== #


class BaseQuantitativeDimension(BaseDimension):
    """A BaseQuantitativeDimension class."""

    __slots__ = (
        "_coordinates_offset",
        "_origin_offset",
        "_quantity_name",
        "_period",
        "_unit",
        "_equivalent_unit",
        "_equivalencies",
    )

    def __init__(
        self,
        description="",
        application={},
        coordinates_offset=None,
        origin_offset=None,
        quantity_name=None,
        period=None,
        label="",
        unit=None,
        **kwargs,
    ):
        """Instantiate a BaseIndependentVariable class."""

        super().__init__(label, application, description)

        self._coordinates_offset = ScalarQuantity(coordinates_offset, unit).quantity
        self._origin_offset = ScalarQuantity(origin_offset, unit).quantity
        self._quantity_name = check_quantity_name(quantity_name, unit)

        value = ScalarQuantity(period, unit).quantity
        self._period = np.inf * value.unit if value.value == 0.0 else value

        self._unit = unit
        self._equivalent_unit = None
        self._equivalencies = None

    def __eq__(self, other):
        check = [getattr(self, _) == getattr(other, _) for _ in __class__.__slots__]
        check += [super().__eq__(other)]
        return np.all(check)

    # ----------------------------------------------------------------------- #
    #                                Attributes                               #
    # ----------------------------------------------------------------------- #

    @property
    def coordinates_offset(self):
        """Value at index zero, :math:`c_k`, along the dimension."""
        return deepcopy(self._coordinates_offset)

    @coordinates_offset.setter
    def coordinates_offset(self, value):
        value = validate(value, "coordinates_offset", ALLOWED_TYPES)
        self._coordinates_offset = ScalarQuantity(value, self._unit).quantity

    @property
    def origin_offset(self):
        """Origin offset, :math:`o_k`, along the dimension."""
        return deepcopy(self._origin_offset)

    @origin_offset.setter
    def origin_offset(self, value):
        value = validate(value, "origin_offset", ALLOWED_TYPES)
        self._origin_offset = ScalarQuantity(value, self._unit).quantity

    @property
    def period(self):
        """Period of the data along this dimension."""
        return deepcopy(self._period)

    @period.setter
    def period(self, value=None):
        value = str(value) if isinstance(value, Quantity) else value
        if not isinstance(value, str):
            raise TypeError(type_error(str, "period", value))

        lst = ["inf", "Inf", "infinity", "Infinity", "∞"]
        is_inf = value.strip().split()[0] in lst
        self._period = (
            np.inf * self._unit
            if is_inf
            else ScalarQuantity(value, self._unit).quantity
        )

    @property
    def quantity_name(self):
        """Quantity name associated with this dimension."""
        return deepcopy(self._quantity_name)

    @quantity_name.setter
    def quantity_name(self, value):
        raise NotImplementedError("This attribute is not yet implemented.")

    @property
    def absolute_coordinates(self):
        """Return the absolute coordinates along the dimensions."""
        return (self.coordinates + self.origin_offset).to(self._unit)

    @property
    def axis_label(self):
        """Return a formatted string for displaying label along the dimension axis."""
        label = self.quantity_name if self.label.strip() == "" else self.label
        unit = self._equivalent_unit or self._unit
        return _axis_label(label, unit)

    # ----------------------------------------------------------------------- #
    #                                  Methods                                #
    # ----------------------------------------------------------------------- #

    def dict(self):
        """Return the object as a python dictionary."""
        obj = {}

        if self._coordinates_offset.value != 0.0:
            obj["coordinates_offset"] = str(ScalarQuantity(self._coordinates_offset))

        if self._origin_offset.value != 0.0:
            obj["origin_offset"] = str(ScalarQuantity(self._origin_offset))

        if self._quantity_name not in [None, "unknown", "dimensionless"]:
            obj["quantity_name"] = self._quantity_name

        if self._period.value not in [0.0, np.inf]:
            obj["period"] = str(ScalarQuantity(self._period))

        obj.update(super().dict())
        return obj

    def is_quantitative(self):
        """Return `True`, if the dimension is quantitative, otherwise `False`.
        :returns: A Boolean.
        """
        return True

    def to(self, unit="", equivalencies=None):
        """Convert the unit to given value `unit`."""
        unit = validate(unit, "unit", str)
        if equivalencies is None:
            self._unit = ScalarQuantity(unit, self._unit).quantity.unit
            self._equivalent_unit = None
            self._equivalencies = None
            return

        # self._unit = ScalarQuantity(unit).quantity.unit
        self._equivalent_unit = ScalarQuantity(unit).quantity.unit
        self._equivalencies = equivalencies


# =========================================================================== #
#                           ReciprocalDimension Class                         #
# =========================================================================== #


class ReciprocalDimension(BaseQuantitativeDimension):
    """ReciprocalDimension class."""



# -*- coding: utf-8 -*-
from os import path

import pytest

import csdmpy as cp

COMMON_PATH = path.join("tests", "file_read", "test_files")


def file_testing(file_, error, error_type):
    with pytest.raises(error_type, match=f".*{error}.*"):
        cp.load(file_)


def test_01():
    file_ = path.join(COMMON_PATH, "1.csdfe")
    error = "'CSDM' is not a valid keyword for the CSD model."
    file_testing(file_, error, KeyError)


def test_02():
    file_ = path.join(COMMON_PATH, "2.csdfe")
    error = "Missing a required `version` key from the CSDM object."
    file_testing(file_, error, KeyError)


def test_03():
    file_ = path.join(COMMON_PATH, "3.csdfe")
    error = "Expecting an instance of type `int` for count, got `str`."
    file_testing(file_, error, TypeError)


def test_04():
    file_ = path.join(COMMON_PATH, "4.csdfe")
    error = "The value, `tensor`, is an invalid `quantity_type`"
    file_testing(file_, error, ValueError)


def test_05():
    file_ = path.join(COMMON_PATH, "5.csdfe")
    error = "The quantity_type, 'vector_2', requires exactly 2 component"
    file_testing(file_, error, Exception)


def test_06():
    file_ = path.join(COMMON_PATH, "6.csdfe")
    error = "The value, `float16`, is an invalid `numeric_type`"
    file_testing(file_, error, ValueError)


def test_07():
    file_ = path.join(COMMON_PATH, "7.csdfe")
    error = "Missing a required `coordinates` key from the MonotonicDimension"
    file_testing(file_, error, KeyError)


def test_08():
    file_ = path.join(COMMON_PATH, "8.csdfe")
    error = "A list of string labels are required, found int"
    file_testing(file_, error, ValueError)


def test_09():
    file_ = path.join(COMMON_PATH, "9.csdfe")
    error = "Missing a required `quantity_type` key from the DependentVariable"
    file_testing(file_, error, KeyError)


def test_10():
    file_ = path.join(COMMON_PATH, "10.csdfe")
    error = "Missing a required `numeric_type` key from the DependentVariable"
    file_testing(file_, error, KeyError)


def test_11():
    file_ = path.join(COMMON_PATH, "11.csdfe")
    error = "The `encoding` key is invalid for DependentVariable objects"
    file_testing(file_, error, KeyError)


def test_12():
    file_ = path.join(COMMON_PATH, "12.csdfe")
    error = "Missing a required 'type' key from the Dimension object."
    file_testing(file_, error, KeyError)


def test_13():
    file_ = path.join(COMMON_PATH, "13.csdfe")
    error = "Missing a required `count` key from the LinearDimension."
    file_testing(file_, error, KeyError)


def test_14():
    file_ = path.join(COMMON_PATH, "14.csdfe")
    error = "Missing a required `type` key from the DependentVariable."
    file_testing(file_, error, KeyError)


def test_15():
    file_ = path.join(COMMON_PATH, "15.csdfe")
    error = "The value, 'blah', is invalid for the `type` attribute"
    file_testing(file_, error, ValueError)


def test_16():
    file_ = path.join(COMMON_PATH, "16.csdfe")
    error = "Missing a required `labels` key from the LabeledDimension object."
    file_testing(file_, error, KeyError)



# -*- coding: utf-8 -*-
"""The SparseSampling class."""
from copy import deepcopy

import numpy as np

from .decoder import Decoder
from csdmpy.utils import check_encoding
from csdmpy.utils import NumericType
from csdmpy.utils import QuantityType
from csdmpy.utils import validate

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["SparseSampling"]


class SparseSampling:
    """Declare a SparseSampling class."""

    __slots__ = (
        "_sparse_dimensions_indexes",
        "_sparse_grid_vertexes",
        "_encoding",
        "_quantity_type",
        "_unsigned_integer_type",
        "_description",
        "_application",
    )

    def __init__(
        self,
        # dimension_indexes,
        # sparse_grid_vertexes,
        # encoding="none",
        # quantity_type="scalar",
        # unsigned_integer_type="int64",
        # description="",
        # application={},
        **kwargs,
    ):
        """Initialize a SparseDimension class."""
        default = {
            "encoding": "none",
            "quantity_type": "scalar",
            "unsigned_integer_type": "uint64",
            "description": "",
            "application": {},
        }
        default.update(kwargs)
        check_sparse_sampling_key_value(default)
        self.encoding = default["encoding"]
        self._unsigned_integer_type = NumericType(default["unsigned_integer_type"])
        self._quantity_type = QuantityType(default["quantity_type"])
        self.description = default["description"]
        self.application = default["application"]
        self._sparse_dimensions_indexes = default["dimension_indexes"]
        self._sparse_grid_vertexes = Decoder(
            self._encoding,
            self._quantity_type,
            [default["sparse_grid_vertexes"]],
            self._unsigned_integer_type.dtype,
        )

    def __eq__(self, other):
        """Overrides the default implementation"""
        if not isinstance(other, SparseSampling):
            return False
        check = [
            np.all(self._sparse_dimensions_indexes == other._sparse_dimensions_indexes),
            np.all(self._sparse_grid_vertexes == other._sparse_grid_vertexes),
            *[getattr(self, _) == getattr(other, _) for _ in __class__.__slots__[2:]],
        ]
        return np.all(check)

    # ----------------------------------------------------------------------- #
    #                                 Attributes                              #
    # ----------------------------------------------------------------------- #

    @property
    def encoding(self):
        """Return the data encoding method."""
        return deepcopy(self._encoding)

    @encoding.setter
    def encoding(self, value):
        self._encoding = validate(value, "encoding", str, check_encoding)

    @property
    def unsigned_integer_type(self):
        """Return the unsigned integer type of data values."""
        return deepcopy(self._unsigned_integer_type)

    @unsigned_integer_type.setter
    def unsigned_integer_type(self, value):
        self._unsigned_integer_type.update(value)

    @property
    def application(self):
        """Return an application metadata dictionary."""
        return deepcopy(self._application)

    @application.setter
    def application(self, value):
        self._application = validate(value, "application", dict)

    @property
    def description(self):
        """Return the description of the object."""
        return deepcopy(self._description)

    @description.setter
    def description(self, value):
        self._description = validate(value, "description", str)

    @property
    def dimension_indexes(self):
        """List of dimension indexes corresponding to sparse dimensions."""
        return deepcopy(self._sparse_dimensions_indexes)

    @property
    def sparse_grid_vertexes(self):
        """List of grid vertexes corresponding to sparse dimensions."""
        return deepcopy(self._sparse_grid_vertexes)


def check_sparse_sampling_key_value(input_dict):
    def message(item):
        return (
            f"Missing a required `{item}` key from the SparseSampling object of the "
            "DependentVariable object."
        )

    sparse_keys = input_dict.keys()
    if "dimension_indexes" not in sparse_keys:
        raise KeyError(message("dimension_indexes"))
    if "sparse_grid_vertexes" not in sparse_keys:
        raise KeyError(message("sparse_grid_vertexes"))
    if "encoding" in sparse_keys:
        code = input_dict["encoding"]
        if code != "none" and "unsigned_integer_type" not in sparse_keys:
            raise KeyError(message("unsigned_integer_type"))

        uint_value = input_dict["unsigned_integer_type"]
        if uint_value not in ["uint8", "uint16", "uint32", "uint64"]:
            raise ValueError(
                f"{uint_value} is an invalid `unsigned_integer_type` enumeration ",
                "literal. The allowed values are `uint8`, `uint16`, `uint32`, ",
                "and `uint64`.",
            )



# -*- coding: utf-8 -*-
import numpy as np

import csdmpy as cp


def setup_reference_offset(offset):
    a = cp.LinearDimension(
        count=6, increment="1", coordinates_offset=f"{offset}", complex_fft=True
    )
    array = np.asarray([-3, -2, -1, 0, 1, 2]) + offset
    assert np.allclose(a.coordinates, array)
    assert a.get_nmr_reference_offset() == offset

    b = cp.LinearDimension(count=6, increment="1", coordinates_offset=f"{-3 + offset}")
    assert np.allclose(b.coordinates, array)
    assert b.get_nmr_reference_offset() == offset

    b = cp.LinearDimension(count=6, increment="-1", coordinates_offset=f"{2+offset}")
    assert np.allclose(b.coordinates, np.asarray([2, 1, 0, -1, -2, -3]) + offset)
    assert b.get_nmr_reference_offset() == offset

    b = cp.LinearDimension(count=5, increment="-1", coordinates_offset=f"{2 + offset}")
    assert np.allclose(b.coordinates, np.asarray([2, 1, 0, -1, -2]) + offset)
    assert b.get_nmr_reference_offset() == offset

    b = cp.LinearDimension(count=5, increment="1", coordinates_offset=f"{-2 + offset}")
    assert np.allclose(b.coordinates, np.asarray([-2, -1, 0, 1, 2]) + offset)
    assert b.get_nmr_reference_offset() == offset


def test_00():
    setup_reference_offset(0)
    setup_reference_offset(10)
    setup_reference_offset(-12.12)
    setup_reference_offset(61.12)
